/*
 * (C) 1996 by Marcin Dalecki <sub994.sub.uni-goettingen.de>
 *
 * This file is part of the Motif Info browser.
 *
 * The GNU Info widget is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <Xm/MainW.h>
#include <Xm/PanedW.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/PushBG.h>
#include <Xm/LabelG.h>
#include <Xm/ScrolledW.h>
#include <Xm/List.h>
#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>

#include <stdio.h>
#include <string.h>

#include "parse.h"
#include "display.h"
#include "file_ops.h"

Widget node_text;
Widget location_text;
Widget back_button;
Widget fwrd_button;
Widget prev_button;
Widget top_button;
Widget next_button;
Widget home_button;
Widget search_button;
Widget top_menu_button;
Widget next_menu_button;
Widget prev_menu_button;
Widget home_menu_button;
Widget menu_item_button;
Widget menu_node_button;
Widget xref_item_button;
Widget xref_node_button;

struct page *current_page;
struct page_stack *page_stack;
struct page_stack *fwrd_stack;


void display_page(struct page *page)
{
    char *buf;
    char *token;
    char label[512];
    char *s[10];
    int i, n;

    /* Update the sesitivity of the all the navigation buttons.
     */

    if (page->Prev) {
	XtVaSetValues(prev_button, XmNsensitive, True, NULL);
	XtVaSetValues(prev_menu_button, XmNsensitive, True, NULL);
    } else {
	XtVaSetValues(prev_button, XmNsensitive, False, NULL);
	XtVaSetValues(prev_menu_button, XmNsensitive, False, NULL);
    }
    if (page->Up) {
	XtVaSetValues(top_button, XmNsensitive, True, NULL);
	XtVaSetValues(top_menu_button, XmNsensitive, True, NULL);
    } else {
	XtVaSetValues(top_button, XmNsensitive, False, NULL);
	XtVaSetValues(top_menu_button, XmNsensitive, False, NULL);
    }
    if (page->Next) {
	XtVaSetValues(next_button, XmNsensitive, True, NULL);
	XtVaSetValues(next_menu_button, XmNsensitive, True, NULL);
    } else {
	XtVaSetValues(next_button, XmNsensitive, False, NULL);
	XtVaSetValues(next_menu_button, XmNsensitive, False, NULL);
    }

    if (page->menu) {
	XtVaSetValues(menu_item_button, XmNsensitive, True, NULL);
	XtVaSetValues(menu_node_button, XmNsensitive, True, NULL);
    } else {
	XtVaSetValues(menu_item_button, XmNsensitive, False, NULL);
	XtVaSetValues(menu_node_button, XmNsensitive, False, NULL);
    }

    if (page->note) {
	XtVaSetValues(xref_item_button, XmNsensitive, True, NULL);
	XtVaSetValues(xref_node_button, XmNsensitive, True, NULL);
    } else {
	XtVaSetValues(xref_item_button, XmNsensitive, False, NULL);
	XtVaSetValues(xref_node_button, XmNsensitive, False, NULL);
    }

    /*
     * Inform the user where we are currently
     */

    label[0] = '\0';
    n = 0;
    if (page->File) {
	s[n++] = strdup("File ");
	s[n++] = strdup(page->File);
    }
    if (page->Node) {
	s[n++] = strdup(", Node ");
	s[n++] = strdup(page->Node);
    }
    if (page->Next) {
	s[n++] = strdup(", Next ");
	s[n++] = strdup(page->Next);
    }
    if (page->Prev) {
	s[n++] = strdup(", Prev ");
	s[n++] = strdup(page->Prev);
    }
    if (page->Up) {
	s[n++] = strdup(", Up ");
	s[n++] = strdup(page->Up);
    }
    for (i = 0; i < n; ++i) {
	strcat(label, s[i]);
	free(s[i]);
    }

    XmTextFieldSetString(location_text, label);

    /* Make sure it doesn't try anything cute until we're ready 
     */
    XmTextDisableRedisplay(node_text);


    XmTextClearSelection(node_text,
			 XtLastTimestampProcessed(XtDisplay(top_level)));
    XmTextSetString(node_text, page->text);

    /* Go for redisplay */
    XmTextEnableRedisplay(node_text);

    XmTextSetInsertionPosition(node_text, 0);

    /* Update the page stack.
     */

    if (current_page) {
	struct page_stack *stack;

	stack = (struct page_stack *) malloc(sizeof(struct page_stack));
	stack->page = current_page;
	stack->prev = page_stack;
	page_stack = stack;
	XtVaSetValues(back_button, XmNsensitive, True, NULL);
    }
    current_page = page;
}
