
/*
 * (C) 1996 by Marcin Dalecki <sub994.sub.uni-goettingen.de>
 *
 * This file is part of the Motif Info browser.
 *
 * The GNU Info widget is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "file_ops.h"
#include "parse.h"
#include "display.h"

#include <alloca.h>

char *lookup_file_in_path(char *name, char *path)
{
    char *buf = alloca(strlen(path) + 1);
    char *token;
    struct stat statb;
    FILE *fp;

    strcpy(buf, path);

    for (token = strtok(buf, ":"); token; token = strtok(NULL, ":")) {
	char *buf = malloc(strlen(token) + strlen(name) + 1);
	int len;

	strcpy(buf, token);
	len = strlen(buf);
	if (len > 0) {
	    if (buf[strlen(buf) - 1] != '/')
		strcat(buf, "/");
	} else
	    strcat(buf, "/");
	strcat(buf, name);
	if (stat(buf, &statb) == -1 || (statb.st_mode & S_IFMT) != S_IFREG ||
	    !(fp = fopen(buf, "r"))) {
	    FILE *input;
	    /*
	     * Try it with a compressed file.
	     */
	    char command[2048];
	    char *readbuf;
	    int size;
	    int readed;
	    int tmp;

	    strcpy(command, "zcat ");
	    strcat(command, buf);
	    strcat(command, " 2> /dev/null");

	    if ((input = popen(command, "r"))) {
		readed = 0;
		size = 0;
		readbuf = (char *) malloc(4096 * sizeof(char) + 2);
		while (tmp = fread(readbuf + size, sizeof(char), 4096, input)) {
		    readed += tmp;
		    size += 4096;
		    readbuf = (char *) realloc(readbuf, 4096 + size + 2);
		}
		if (readed) {
		    char *text;

		    readbuf[readed] = '\0';
		    text = strdup(readbuf);
		    free(readbuf);
		    return text;
		}
		free(readbuf);
	    }
	    free(buf);
	} else {
	    char *text = NULL;

	    /* put the contents of the file in the Text widget by allocating
	     * enough space for the entire file, reading the file into the
	     * allocated space, and using XmTextFieldSetString() to show the file.
	     */
	    if (!(text = (char *) malloc((unsigned) (statb.st_size + 1)))) {
		fprintf(stderr, "Can't alloc enough space for %s\n", buf);
		free(buf);
		fclose(fp);
	    } else {
		if (!fread(text, sizeof(char), statb.st_size + 1, fp)) {
		    fprintf(stderr, "Warning: may not have read entire file!\n");
		}
		text[statb.st_size] = '\0';	/* be sure to NULL-terminate */

		free(buf);

		return text;
	    }
	}
    }

    return NULL;
}


char *file_path;

struct file_cache {
    struct file_cache *next;
    char *file;
    struct top *top;		/* not zero if it is an indirection table */
    struct page *page;		/* of course every file has an contents */
} *file_cache = NULL;

static struct page *lookup_node_in_file_rr(int rr, char *node, char *file)
{
    struct file_cache *cache;
    struct page *page;
    char *tmp;
    char *name_dup;

    /*
     * Bump out if we are ging to call ourselfs in infinite recursion!
     * This helps with broken info files, containing not present refferences.
     */

    if (rr <= 0)
	return NULL;

    /* Handle the case of file indirect nodes.
     * Currently only for the DIR node.
     */

    if (!strcmp(node, "(dir)") || !strcmp(node, "(DIR)")) {
	node = "Top";
	file = "dir";
    }
    /* First look if the file is allready cached.
     */

    cache = file_cache;
    while (cache && strcmp(cache->file, file))
	cache = cache->next;

    /* If not allready found there, try to look it up and then add it to
     * the cache, return NULL otherwise.
     */
    if (!cache) {
	char *text;

	text = lookup_file_in_path(file, file_path);
	if (!text)
	    return NULL;

	cache = (struct file_cache *) malloc(sizeof(struct file_cache));
	cache->top = parse_top(text);
#if 0
	debug_print_top(cache->top);
#endif
	cache->file = strdup(file);
	if (!cache->top) {
	    struct page *page;
	    cache->page = parse_file(file, text);
	    page = cache->page;
#if 0
	    while (page) {
		debug_print_page(page);
		page = page->next;
	    }
#endif
	} else
	    cache->page = NULL;

	/* Add it to the cache.
	 */
	cache->next = file_cache;
	file_cache = cache;
    }
    /*
     * Try it in the current file from the cache.
     */
    page = cache->page;

    while (page && page->Node && strcmp(page->Node, node)) {
	page = page->next;
    }

    if (page && page->Node)
	return page;

    /* There is an associated indirection table try it there.
     */
    if (cache->top) {
	struct node *temp;

	for (temp = cache->top->node; temp; temp = temp->next) {
	    if (!strcmp(temp->node, node))
		return lookup_node_in_file_rr(rr - 1, node, temp->file);
	}
    }
    /* Last resort, try to find the indirection table.
     */
    name_dup = strdup(cache->file);
    if ((tmp = strstr(name_dup, ".info-"))) {
	tmp[5] = '\0';
	page = lookup_node_in_file_rr(rr - 1, node, name_dup);
	free(name_dup);
	return page;
    } else if ((tmp = strstr(name_dup, "info-"))) {
	tmp[4] = '\0';
	page = lookup_node_in_file_rr(rr - 1, node, name_dup);
	free(name_dup);
	return page;
    } else {
	name_dup = (char *) realloc(name_dup, strlen(name_dup) + 6);
	strcat(name_dup, ".info");
	page = lookup_node_in_file_rr(rr - 1, node, name_dup);
	free(name_dup);
	return page;
    }
    free(name_dup);

    return NULL;
}

struct page *lookup_node_in_file(char *node, char *file)
{
    return lookup_node_in_file_rr(100, node, file);
}

static void load_file(dialog, client_data, call_data)
Widget dialog;
XtPointer client_data;
XtPointer call_data;
{
    Pixmap pixmap;
    char *file = NULL;
    char *file_dup;
    struct page *page;
    char *current_path;
    char *used_path;
    char *saved_path;
    char *trunc;

    XmFileSelectionBoxCallbackStruct *cbs =
    (XmFileSelectionBoxCallbackStruct *) call_data;

    if (!cbs)
	return;

    if (!XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &file))
	return;			/* internal error */
    XtPopdown(XtParent(dialog));

    /*
     * Preprocess the filename.
     * 1. Remove the compression endings .gz .Z .z
     * 2. Remove the leading path and save it
     */

    file_dup = current_path = file;
    trunc = strrchr(file, '.');
    if (trunc && (!strcmp(trunc, ".gz") || !strcmp(trunc, ".Z")
		  || !strcmp(trunc, ".z"))) {
	/* Remove compression marks they are only confusing.
	 * We are handling them "automagically" anyway.
	 */
	*trunc = '\0';
    }
    saved_path = file_path;
    if (trunc = strrchr(file, '/')) {
	file = trunc + 1;
	*trunc = '\0';
    } else {
	current_path = ".";
    }
    used_path = (char *) malloc(strlen(current_path) + 1
				+ strlen(file_path) + 1 + 2);

    strcpy(used_path, current_path);
    strcat(used_path, "/:");
    strcat(used_path, file_path);

    /* Communicate the new extended file search path.
     */
    file_path = used_path;
    page = lookup_node_in_file("Top", file);
    if (page) {
	display_page(page);
	/*
	 * Add the path to the file search path! (Actually it's allready done).
	 */
	file_path = used_path;
	free(saved_path);
    } else {
	free(used_path);
	file_path = saved_path;

	/* Inform the user about what happaned!
	 */

	error_dialog("File not found, or it doesn't\n"
		     "contain any Top info node!");
    }
    XtFree(file_dup);
}

/* Any item the user selects from the File menu calls this function.
 * It will either be "Open" (item_no == 0) or "Quit" (item_no == 1).
 */
void file_cb(widget, client_data, call_data)
Widget widget;			/* menu item that was selected */
XtPointer client_data;		/* the index into the menu */
XtPointer call_data;		/* unused */
{
    static Widget dialog;	/* make it static for reuse */
    int item_no = (int) client_data;

    if (item_no == 1)		/* the "quit" item */
	exit(0);

    /* "Open" was selected.  Create a Motif FileSelectionDialog w/callback */
    if (!dialog) {
	dialog = (Widget) XmCreateFileSelectionDialog(top_level,
						    "file_sel", NULL, 0);
	XtAddCallback(dialog, XmNokCallback, load_file, NULL);
	XtAddCallback(dialog, XmNcancelCallback,
		      (XtCallbackProc) XtUnmanageChild, NULL);

	XtUnmanageChild((void *) XmFileSelectionBoxGetChild(dialog,
						  XmDIALOG_HELP_BUTTON));

    }
#if 0
    XtVaSetValues(dialog, XmNfileFilterStyle, XmFILTER_HIDDEN_FILES,
		  XmNwidth, 500,
		  NULL);
#endif
    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);
}
