
/*
 * (C) 1996 by Marcin Dalecki <sub994.sub.uni-goettingen.de>
 *
 * This file is part of the Motif Info browser.
 *
 * The GNU Info widget is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <Xm/MainW.h>
#include <Xm/PanedW.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/RowColumn.h>
#include <Xm/PushBG.h>
#include <Xm/SeparatoG.h>
#include <Xm/ScrolledW.h>
#include <Xm/List.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>

#include <X11/xpm.h>

#include <stdio.h>

#include "pixmaps/icon.xpm"
#include "pixmaps/back_arm.xpm"
#include "pixmaps/back_dis.xpm"
#include "pixmaps/fwrd_arm.xpm"
#include "pixmaps/fwrd_dis.xpm"
#include "pixmaps/prev_arm.xpm"
#include "pixmaps/prev_dis.xpm"
#include "pixmaps/next_arm.xpm"
#include "pixmaps/next_dis.xpm"
#include "pixmaps/top_arm.xpm"
#include "pixmaps/top_dis.xpm"
#include "pixmaps/home.xpm"
#include "pixmaps/search.xpm"
#include "pixmaps/dalecki.xpm"

#include "file_ops.h"
#include "display.h"
#include "parse.h"
#include "utils.h"

typedef struct {
    String path;
    String file;
    String node;
} app_data_t;

app_data_t app_data;

static String fallback_resources[] =
{
    "*fontList: -*-helvetica-bold-r-*-*-*-120-*-*-*-*-iso8859-*",
"*XmTextField.fontList: -*-courier-medium-r-*-*-*-120-*-*-*-*-iso8859-*",
    "*XmText.fontList: fixed",
    "*XmList*fontList: -*-courier-medium-r-*-*-*-120-*-*-*-*-iso8859-*",
    "*foreground: Black",
    "*background: #B2B2B2",
    "*XmText.foreground: Black",
    "*XmText.background: #C0C0C0",
    "*borderColor: #B2B2B2",
    "*XmTextField.foreground: Black",
    "*XmTextField.background: #D3B5B5",
    "*XmList.foreground: Black",
    "*XmList.background: #D3B5B5",
    NULL
};

static XtResource new_resources[] =
{
    {"dirPath", "DirPath", XmRString, sizeof(String),
     XtOffsetOf(app_data_t, path), XmRString,
     (XtPointer) "/opt/info/"},
    {"dirFile", "DirFile", XmRString, sizeof(String),
     XtOffsetOf(app_data_t, file), XmRString,
     (XtPointer) "dir"},
    {"topNode", "topNode", XmRString, sizeof(String),
     XtOffsetOf(app_data_t, node), XmRString,
     (XtPointer) "Top"},
    NULL,
};


static XrmOptionDescRec options[] =
{
    {"-file", ".dirFile", XrmoptionSepArg, NULL},
    {"-path", ".dirPath", XrmoptionSepArg, NULL},
};

Widget top_level;

static void home_pushed(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    struct page *page = lookup_node_in_file("Top", "dir");
    if (page)
	display_page(page);
    else
	error_dialog("Info directory file not found!");
}

static void next_pushed(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    struct page *page;

    page = NULL;

    if (current_page)
	page = lookup_node_in_file(current_page->Next, current_page->file);
    if (page)
	display_page(page);
    else
	error_dialog("Next node not found!");
}

static void prev_pushed(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    struct page *page;

    page = NULL;

    if (current_page)
	page = lookup_node_in_file(current_page->Prev, current_page->file);
    if (page)
	display_page(page);
    else
	error_dialog("Preview node not found!");
}

static void top_pushed(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    struct page *page;

    page = NULL;

    if (current_page)
	page = lookup_node_in_file(current_page->Up, current_page->file);
    if (page)
	display_page(page);
    else
	error_dialog("Up node not found!");
}

static void back_pushed(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    struct page_stack *tmp;
    struct page_stack *fwrd;

    if (!page_stack)
	return;

    if (!page_stack->prev)
	XtVaSetValues(back_button, XmNsensitive, False, NULL);
    if (current_page) {
	fwrd = (struct page_stack *) malloc(sizeof(struct page_stack));
	fwrd->prev = fwrd_stack;
	fwrd->page = current_page;
	fwrd_stack = fwrd;
    }
    current_page = NULL;
    display_page(page_stack->page);
    tmp = page_stack;
    page_stack = page_stack->prev;
    free(tmp);
    XtVaSetValues(fwrd_button, XmNsensitive, True, NULL);
}

static void fwrd_pushed(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    struct page_stack *tmp;

    if (!fwrd_stack)
	return;
    if (!fwrd_stack->prev)
	XtVaSetValues(fwrd_button, XmNsensitive, False, NULL);
    current_page = NULL;
    display_page(fwrd_stack->page);
    tmp = fwrd_stack;
    fwrd_stack = fwrd_stack->prev;
    free(tmp);
}

static void load_menu_by_name(dialog, client_data, call_data)
Widget dialog;
XtPointer client_data;
XtPointer call_data;
{
    Pixmap pixmap;
    char *menu_item = NULL;
    XmFileSelectionBoxCallbackStruct *cbs =
    (XmFileSelectionBoxCallbackStruct *) call_data;

    if (cbs) {
	struct menu *menu;

	XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &menu_item);
	menu = current_page->menu;
	while (menu && strcmp(menu->name, menu_item))
	    menu = menu->next;
	XtFree(menu_item);
	if (menu) {
	    char *node;
	    char *file;
	    struct page *page;

	    page = NULL;
	    node = file = NULL;

	    if (*menu->node == '(') {	/* File indirection */
		node = "Top";
		file = strdup(menu->node + 1);
		if (strlen(file))
		    file[strlen(file) - 1] = '\0';
	    } else {
		node = menu->node;
		file = strdup(current_page->file);
	    }

	    page = lookup_node_in_file(node, file);
	    free(file);

	    if (page)
		display_page(page);
	    else
		error_dialog("Menu item not found!");

	    XtPopdown(XtParent(dialog));
	} else {
	    error_dialog("Menu item not found!");

	    XtPopdown(XtParent(dialog));
	}
    }
}

static void load_menu_by_node(dialog, client_data, call_data)
Widget dialog;
XtPointer client_data;
XtPointer call_data;
{
    Pixmap pixmap;
    char *menu_node = NULL;
    XmFileSelectionBoxCallbackStruct *cbs =
    (XmFileSelectionBoxCallbackStruct *) call_data;

    if (cbs) {
	struct menu *menu;

	XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &menu_node);
	menu = current_page->menu;
	while (menu && strcmp(menu->node, menu_node))
	    menu = menu->next;
	XtFree(menu_node);
	if (menu) {
	    char *node;
	    char *file;
	    struct page *page;

	    page = NULL;
	    node = file = NULL;

	    if (*menu->node == '(') {	/* File indirection */
		node = "Top";
		file = strdup(menu->node + 1);
		if (strlen(file))
		    file[strlen(file) - 1] = '\0';

	    } else {
		node = menu->node;
		file = strdup(current_page->file);
	    }

	    page = lookup_node_in_file(node, file);
	    free(file);

	    if (page)
		display_page(page);
	    else
		error_dialog("Menu node not found!");

	    XtPopdown(XtParent(dialog));
	} else
	    error_dialog("Menu item not found!");
    }
}

static void go_cb(widget, client_data, call_data)
Widget widget;			/* menu item that was selected */
XtPointer client_data;		/* the index into the menu */
XtPointer call_data;		/* unused */
{
    int item_no = (int) client_data;

    if (item_no == 0)
	next_pushed(widget, client_data, call_data);
    if (item_no == 1)
	prev_pushed(widget, client_data, call_data);
    if (item_no == 2)
	top_pushed(widget, client_data, call_data);
    if (item_no == 5)
	home_pushed(widget, client_data, call_data);
}

static void go_menu_cb(widget, client_data, call_data)
Widget widget;			/* menu item that was selected */
XtPointer client_data;		/* the index into the menu */
XtPointer call_data;		/* unused */
{
    int item_no = (int) client_data;

    if (item_no == 0 || item_no == 1) {		/* Go to menu item/node */
	Widget dialog;
	int n;
	int i;
	XmString *str;
	XmString t;

	if (current_page) {
	    struct menu *menu = current_page->menu;
	    if (menu) {
		n = 0;
		while (menu) {
		    menu = menu->next;
		    n++;
		}

		str = (XmString *) XtMalloc(n * sizeof(XmString));
		if (item_no == 0)
		    t = XmStringCreateLocalized("Please select an menu item:");
		else
		    t = XmStringCreateLocalized("Please select an menu node:");
		menu = current_page->menu;
		for (i = 0; i < n; i++) {
		    if (item_no == 0)
			str[i] = XmStringCreateLocalized(menu->name);
		    else
			str[i] = XmStringCreateLocalized(menu->node);
		    menu = menu->next;
		}

		dialog = (Widget) XmCreateSelectionDialog(top_level,
						     "menuSel", NULL, 0);
		XtUnmanageChild((void *) XmSelectionBoxGetChild(dialog,
						  XmDIALOG_HELP_BUTTON));
		XtUnmanageChild((void *) XmSelectionBoxGetChild(dialog,
						 XmDIALOG_APPLY_BUTTON));
		XtVaSetValues(dialog,
			      XmNlistLabelString, t,
			      XmNlistItems, str,
			      XmNlistItemCount, n,
			      XmNmustMatch, True,
			      NULL);

		if (item_no == 0)
		    XtAddCallback(dialog, XmNokCallback,
				  load_menu_by_name, NULL);
		else
		    XtAddCallback(dialog, XmNokCallback,
				  load_menu_by_node, NULL);

		XtAddCallback(dialog, XmNcancelCallback,
			      (XtCallbackProc) XtDestroyWidget, NULL);
		XmStringFree(t);
		while (--i >= 0)
		    XmStringFree(str[i]);	/* free elements of array */
		XtFree((void *) str);	/* now free array pointer */

		XtManageChild(dialog);
		XtPopup(XtParent(dialog), XtGrabNone);
	    }
	} else
	    error_dialog("There isn't any node currently loaded!");
    }
}


static void load_xref_by_name(dialog, client_data, call_data)
Widget dialog;
XtPointer client_data;
XtPointer call_data;
{
    Pixmap pixmap;
    char *xref_item = NULL;
    XmFileSelectionBoxCallbackStruct *cbs =
    (XmFileSelectionBoxCallbackStruct *) call_data;

    if (cbs) {
	struct menu *xref;

	XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &xref_item);
	xref = current_page->note;
	while (xref && strcmp(xref->name, xref_item))
	    xref = xref->next;
	XtFree(xref_item);
	if (xref) {
	    char *node;
	    char *file;
	    struct page *page;

	    page = NULL;
	    node = file = NULL;

	    if (*xref->node == '(') {	/* File indirection */
		node = "Top";
		file = strdup(xref->node + 1);
		if (strlen(file))
		    file[strlen(file) - 1] = '\0';

	    } else {
		node = xref->node;
		file = strdup(current_page->file);
	    }

	    page = lookup_node_in_file(node, file);
	    free(file);

	    if (page)
		display_page(page);
	    else
		error_dialog("Cross refference not found!");

	    XtPopdown(XtParent(dialog));
	} else
	    error_dialog("Cross refference not found!");
    }
}


static void load_xref_by_node(dialog, client_data, call_data)
Widget dialog;
XtPointer client_data;
XtPointer call_data;
{
    Pixmap pixmap;
    char *xref_node = NULL;
    XmFileSelectionBoxCallbackStruct *cbs =
    (XmFileSelectionBoxCallbackStruct *) call_data;

    if (cbs) {
	struct menu *xref;

	XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &xref_node);
	xref = current_page->note;
	while (xref && strcmp(xref->node, xref_node))
	    xref = xref->next;
	XtFree(xref_node);
	if (xref) {
	    char *node;
	    char *file;
	    struct page *page;

	    page = NULL;
	    node = file = NULL;

	    if (*xref->node == '(') {	/* File indirection */
		node = "Top";
		file = strdup(xref->node + 1);
		if (strlen(file))
		    file[strlen(file) - 1] = '\0';

	    } else {
		node = xref->node;
		file = strdup(current_page->file);
	    }

	    page = lookup_node_in_file(node, file);
	    free(file);

	    if (page)
		display_page(page);
	    else
		error_dialog("Cress refference node not found!");

	    XtPopdown(XtParent(dialog));
	} else
	    error_dialog("Cross refferences not found!");
    }
}

static void go_xref_cb(widget, client_data, call_data)
Widget widget;			/* menu item that was selected */
XtPointer client_data;		/* the index into the menu */
XtPointer call_data;		/* unused */
{
    int item_no = (int) client_data;

    if (item_no == 0 || item_no == 1) {		/* Go to menu item/node */
	Widget dialog;
	int n;
	int i;
	XmString *str;
	XmString t;

	if (current_page) {
	    struct menu *xref = current_page->note;

	    if (xref) {
		n = 0;
		while (xref) {
		    xref = xref->next;
		    n++;
		}

		str = (XmString *) XtMalloc(n * sizeof(XmString));
		if (item_no == 0)
		    t = XmStringCreateLocalized("Please select a refference:");
		else
		    t = XmStringCreateLocalized(
				       "Please select a reffered node:");
		xref = current_page->note;
		for (i = 0; i < n; i++) {
		    if (item_no == 0)
			str[i] = XmStringCreateLocalized(xref->name);
		    else
			str[i] = XmStringCreateLocalized(xref->node);
		    xref = xref->next;
		}

		dialog = (Widget) XmCreateSelectionDialog(top_level,
						     "xrefSel", NULL, 0);
		XtUnmanageChild((void *) XmSelectionBoxGetChild(dialog,
						  XmDIALOG_HELP_BUTTON));
		XtUnmanageChild((void *) XmSelectionBoxGetChild(dialog,
						 XmDIALOG_APPLY_BUTTON));
		XtVaSetValues(dialog,
			      XmNlistLabelString, t,
			      XmNlistItems, str,
			      XmNlistItemCount, n,
			      XmNmustMatch, True,
			      NULL);

		if (item_no == 0)
		    XtAddCallback(dialog, XmNokCallback,
				  load_xref_by_name, NULL);
		else
		    XtAddCallback(dialog, XmNokCallback,
				  load_xref_by_node, NULL);

		XtAddCallback(dialog, XmNcancelCallback,
			      (XtCallbackProc) XtDestroyWidget, NULL);
		XmStringFree(t);
		while (--i >= 0)
		    XmStringFree(str[i]);	/* free elements of array */
		XtFree((void *) str);	/* now free array pointer */

		XtManageChild(dialog);
		XtPopup(XtParent(dialog), XtGrabNone);
	    }
	} else
	    error_dialog("There isn't any node currently loaded!");
    }
}

#define MSG \
"Use the File menu entry to open different info directories\n\
Use the Go menu entry to navigate through the info files.\n\
Click on menu entries and notes to get through them.\n\
However using the pull down menu selection method is much\n\
more reliable.\n\n\
The two leftmost buttons are used to handle the history list\n\
of allready traversed Nodes\n\n\
Perhaps some time in the future I will find the time to\n\
write something meaningfull at this place."

#define ABOUT \
"This program was written by:\n\n"\
"            Copyright 1996, Marcin Dalecki\n\n"\
"You can reach me currently at the following email\n"\
"address:\n\n"\
"     <dalecki@sub994.sub.uni-goettingen.de>\n\n"\
"The program version is 1.4.1."


/* The help button in the help menu from the menubar was selected.
 * Display help information defined above for how to use the program.
 * This is done by creating a Motif information dialog box.  Again,
 * make the dialog static so we can reuse it.
 */
static void help_cb(widget, client_data, call_data)
Widget widget;
XtPointer client_data;
XtPointer call_data;
{
    Widget dialog;
    int item_no = (int) client_data;
    Arg args[2];
    int i;
    XmString msg;
    Display *dsp;
    Window win;
    static Pixmap image;
    static once = 1;

    dsp = XtDisplay(top_level);
    win = XRootWindowOfScreen(XtScreen(top_level));
    if (once) {
	XpmCreatePixmapFromData(dsp, win, dalecki_xpm, &image, NULL, NULL);
	once = 0;
    }
    if (item_no != 0)
	msg = XmStringCreateLtoR(MSG, XmFONTLIST_DEFAULT_TAG);
    else
	msg = XmStringCreateLtoR(ABOUT, XmFONTLIST_DEFAULT_TAG);

    i = 0;
    XtSetArg(args[i], XmNmessageString, msg);
    i++;
    if (item_no == 0) {
	XtSetArg(args[i], XmNsymbolPixmap, image);
	i++;
    }
    dialog = (Widget) XmCreateInformationDialog(top_level,
						"aboutDialog", args, i);
    XtUnmanageChild((void *) XmMessageBoxGetChild(dialog,
						  XmDIALOG_HELP_BUTTON));
    XtUnmanageChild((void *) XmMessageBoxGetChild(dialog,
						XmDIALOG_CANCEL_BUTTON));
    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);
}

#undef MSG
#undef ABOUT

static void text_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
    XmTextPosition beg, end, nlen, i, posl;
    XmTextVerifyPtr data = (void *) call_data;
    char ret[512];
    char *tmp;

    if (!current_page)
	return;

    if (data->reason != XmCR_MOVING_INSERT_CURSOR)
	return;

    posl = data->newInsert;
    /* Next, try and get a complete "item" */
    if (!XmTextFindString(w, posl, "*", XmTEXT_BACKWARD, &beg))
	return;			/* Bomb out */

    if (posl - beg > 64)	/* just for assecurance */
	return;

    if (XmTextFindString(w, posl, ":", XmTEXT_FORWARD, &end) &&
	end > beg) {
	long len = end - beg;

	/* Victory! Now try and figure out what it is */
	tmp = ret;
	if (!XmTextGetSubstring(w, beg, len, 512, tmp))
	    return;		/* If can't read, forget it */
	else {
	    i = len + 1;
	    while ((*tmp == '*' || isspace(*tmp)) && i) {
		++tmp;
		--i;
	    }
	}
	if (i <= 0)
	    return;		/* Nothing left, forget it */
	else {
	    len = i;
	    for (i = 0; i < len; ++i)
		if (tmp[i] == '\n')
		    tmp[i] = ' ';
	    tmp[i] = '\0';

	    normalize_white_space(&tmp);

	    if ((tmp != strstr(tmp, "note "))
		&& (tmp != strstr(tmp, "Note "))) {
		struct menu *menu;

		/* It was a menu point. Search it there.
		 */
		menu = current_page->menu;
		while (menu && strcmp(menu->name, tmp))
		    menu = menu->next;

		if (menu) {
		    char *node;
		    char *file;
		    struct page *page;

		    page = NULL;
		    node = file = NULL;

		    if (*menu->node == '(') {	/* File indirection */
			node = "Top";
			file = strdup(menu->node + 1);
			if (strlen(file))
			    file[strlen(file) - 1] = '\0';
		    } else {
			node = menu->node;
			file = strdup(current_page->file);
		    }

		    page = lookup_node_in_file(node, file);
		    free(file);
		    if (page) {
			display_page(page);
		    }
		    return;
		}
		/* We don't do anything otherwise, since this would be
		 * too annoying for the user.
		 */

	    } else {
		char *name = strdup(tmp + 5);
		struct page *page;

		if (!XmTextGetSubstring(w, end, 2, 512, tmp)) {
		    free(name);
		    return;	/* If can't read, forget it */
		}
		if (strcmp(tmp, "::")) {
		    fprintf(stderr, "Sorry no cross file notes supported\n");
		    free(name);
		    return;
		}
		page = lookup_node_in_file(name, current_page->file);
		free(name);
		if (page) {
		    display_page(page);
		}
	    }
	}
    }
}

int main(int argc, char **argv)
{
    XtAppContext app;
    Arg args[32];
    Cardinal i;
    Widget widget;
    Widget main_window;
    Widget menu_bar;
    Widget form;
    Widget button_row;
    Widget location_row;
    Widget menu_pullright;
    Widget separator;

    XmString str[32];

    XmString help;

    /* Data needed for Icon creation
     */
    Display *dsp;
    Window win;
    Pixmap shape;
    Pixmap image;
    XpmAttributes attributes;
    XpmColorSymbol symbols[1];
    Pixel color;



    struct page *first_page;

    XtSetLanguageProc(NULL, NULL, NULL);

    top_level = XtVaAppInitialize(&app, "Minfo", options, XtNumber(options),
				  &argc, argv, fallback_resources);

    XtGetApplicationResources(top_level, (XtPointer) & app_data, new_resources,
			      XtNumber(new_resources), NULL, 0);

    main_window = XtVaCreateWidget("main_window",
				   xmMainWindowWidgetClass, top_level,
				   NULL);

    dsp = XtDisplay(top_level);
    win = XRootWindowOfScreen(XtScreen(top_level));

    XtVaGetValues(top_level, XmNbackground, &color, NULL);

    XpmCreatePixmapFromData(dsp, win, icon_xpm, &image, &shape, NULL);
    XtVaSetValues(top_level, XmNiconPixmap, image,
		  XmNiconMask, shape, NULL);

    str[0] = XmStringCreateLocalized("File");
    str[1] = XmStringCreateLocalized("Go");
    help = XmStringCreateLocalized("Help");
    menu_bar = XmVaCreateSimpleMenuBar(main_window, "menubar",
				       XmVaCASCADEBUTTON, str[0], NULL,
				       XmVaCASCADEBUTTON, str[1], NULL,
				       XmVaCASCADEBUTTON, help, NULL,
				       NULL);
    XmStringFree(str[0]);
    XmStringFree(str[1]);

    /* Tell the menubar which button is the help menu  
     */
    if (widget = XtNameToWidget(menu_bar, "button_2"))
	XtVaSetValues(menu_bar, XmNmenuHelpWidget, widget, NULL);

    /* First menu is the File menu -- callback is file_cb() 
     */
    str[0] = XmStringCreateLocalized("Open File...");
    str[1] = XmStringCreateLocalized("Alt+O");
    str[2] = XmStringCreateLocalized("Exit");
    str[3] = XmStringCreateLocalized("Alt+Q");
    XmVaCreateSimplePulldownMenu(menu_bar, "file_menu", 0, file_cb,
		  XmVaPUSHBUTTON, str[0], 'O', "Alt ~Ctrl<Key>q", str[1],
				 XmVaSEPARATOR,
		  XmVaPUSHBUTTON, str[2], 'x', "Alt ~Ctrl<Key>q", str[3],
				 NULL);
    XmStringFree(str[0]);
    XmStringFree(str[1]);
    XmStringFree(str[2]);
    XmStringFree(str[3]);

    /* Second menu is the Go menu -- callback is go_cb() 
     */
    str[0] = XmStringCreateLocalized("Next");
    str[1] = XmStringCreateLocalized("Alt+Right");
    str[2] = XmStringCreateLocalized("Preview");
    str[3] = XmStringCreateLocalized("Alt+Left");
    str[4] = XmStringCreateLocalized("Top");
    str[5] = XmStringCreateLocalized("Alt+Up");
    str[6] = XmStringCreateLocalized("Menu by...");
    str[7] = XmStringCreateLocalized("Cross Refference by...");
    str[8] = XmStringCreateLocalized("Info Directory Node");
    widget = XmVaCreateSimplePulldownMenu(menu_bar, "go_menu", 1, go_cb,
	  XmVaPUSHBUTTON, str[0], NULL, "Alt ~Ctrl<Key>osfRight", str[1],
	   XmVaPUSHBUTTON, str[2], NULL, "Alt ~Ctrl<Key>osfLeft", str[3],
	     XmVaPUSHBUTTON, str[4], NULL, "Alt ~Ctrl<Key>osfUp", str[5],
					  XmVaSEPARATOR,
					  XmVaCASCADEBUTTON, str[6], 'M',
					  XmVaCASCADEBUTTON, str[7], 'R',
					  XmVaSEPARATOR,
				XmVaPUSHBUTTON, str[8], NULL, NULL, NULL,
					  NULL);
    for (i = 0; i <= 8; ++i)
	XmStringFree(str[i]);

    next_menu_button = XtNameToWidget(widget, "button_0");
    prev_menu_button = XtNameToWidget(widget, "button_1");
    top_menu_button = XtNameToWidget(widget, "button_2");

    str[0] = XmStringCreateLocalized("Item");
    str[1] = XmStringCreateLocalized("Node");
    menu_pullright = XmVaCreateSimplePulldownMenu(widget,
		      "pullRight", 4 /* menu item offset */ , go_menu_cb,
				 XmVaPUSHBUTTON, str[0], 'I', NULL, NULL,
				 XmVaPUSHBUTTON, str[1], 'N', NULL, NULL,
						  NULL);

    menu_item_button = XtNameToWidget(menu_pullright, "button_0");
    menu_node_button = XtNameToWidget(menu_pullright, "button_1");

    menu_pullright = XmVaCreateSimplePulldownMenu(widget,
		      "pullRight", 5 /* menu item offset */ , go_xref_cb,
				 XmVaPUSHBUTTON, str[0], 'I', NULL, NULL,
				 XmVaPUSHBUTTON, str[1], 'N', NULL, NULL,
						  NULL);
    xref_item_button = XtNameToWidget(menu_pullright, "button_0");
    xref_node_button = XtNameToWidget(menu_pullright, "button_1");

    XmStringFree(str[0]);
    XmStringFree(str[1]);
    XtManageChild(widget);

    XtVaSetValues(next_menu_button, XmNsensitive, False, NULL);
    XtVaSetValues(prev_menu_button, XmNsensitive, False, NULL);
    XtVaSetValues(top_menu_button, XmNsensitive, False, NULL);
    XtVaSetValues(menu_item_button, XmNsensitive, False, NULL);
    XtVaSetValues(menu_node_button, XmNsensitive, False, NULL);
    XtVaSetValues(xref_item_button, XmNsensitive, False, NULL);
    XtVaSetValues(xref_node_button, XmNsensitive, False, NULL);

    /* Third menu is the help menu -- callback is help_cb() 
     */
    str[0] = XmStringCreateLocalized("About Minfo");
    XmVaCreateSimplePulldownMenu(menu_bar, "help_menu", 2, help_cb,
				 XmVaPUSHBUTTON, str[0], 'A', NULL, NULL,
				 XmVaSEPARATOR,
				 XmVaPUSHBUTTON, help, 'H', NULL, NULL,
				 NULL);
    XmStringFree(str[0]);
    XmStringFree(help);		/* we're done with it; now we can free it */

    XtManageChild(menu_bar);

    form = XtVaCreateWidget("form", xmFormWidgetClass, main_window,
			    XmNorientation, XmVERTICAL,
			    NULL);

    button_row = XtVaCreateWidget("buttonRow", xmRowColumnWidgetClass, form,
				  XmNorientation, XmHORIZONTAL,
				  XmNtopAttachment, XmATTACH_FORM,
				  XmNleftAttachment, XmATTACH_FORM,
				  XmNrightAttachment, XmATTACH_FORM,
				  NULL);

    XtVaCreateManagedWidget(" ", xmLabelWidgetClass, button_row,
			    NULL);

    back_button = XtVaCreateManagedWidget("backButton",
				     xmPushButtonGadgetClass, button_row,
					  XmNsensitive, False,
					  XmNshowAsDefault, False,
					  NULL);
    XtAddCallback(back_button, XmNactivateCallback, back_pushed, NULL);

    XtVaGetValues(back_button, XmNbackground, &color, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;

    XpmCreatePixmapFromData(dsp, win, back_arm, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(back_button, XmNlabelType, XmPIXMAP,
		  XmNarmPixmap, image, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;
    XpmCreatePixmapFromData(dsp, win, back_dis, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(back_button, XmNlabelInsensitivePixmap, image, NULL);

    fwrd_button = XtVaCreateManagedWidget("forwardButton",
				     xmPushButtonGadgetClass, button_row,
					  XmNsensitive, False,
					  XmNshowAsDefault, False,
					  NULL);
    XtAddCallback(fwrd_button, XmNactivateCallback, fwrd_pushed, NULL);

    XtVaGetValues(fwrd_button, XmNbackground, &color, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;

    XpmCreatePixmapFromData(dsp, win, fwrd_arm, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(fwrd_button, XmNlabelType, XmPIXMAP,
		  XmNarmPixmap, image,
		  NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;
    XpmCreatePixmapFromData(dsp, win, fwrd_dis, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(fwrd_button, XmNlabelInsensitivePixmap, image, NULL);

    XtVaCreateManagedWidget(" ", xmLabelWidgetClass, button_row, NULL);

    prev_button = XtVaCreateManagedWidget("prevButton",
				     xmPushButtonGadgetClass, button_row,
					  XmNsensitive, False,
					  XmNshowAsDefault, False,
					  NULL);
    XtAddCallback(prev_button, XmNactivateCallback, prev_pushed, NULL);

    XtVaGetValues(prev_button, XmNbackground, &color, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;

    XpmCreatePixmapFromData(dsp, win, prev_arm, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(prev_button, XmNlabelType, XmPIXMAP,
		  XmNarmPixmap, image,
		  NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;
    XpmCreatePixmapFromData(dsp, win, prev_dis, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(prev_button, XmNlabelInsensitivePixmap, image, NULL);

    top_button = XtVaCreateManagedWidget("topButton",
				     xmPushButtonGadgetClass, button_row,
					 XmNsensitive, False,
					 XmNshowAsDefault, False,
					 NULL);
    XtAddCallback(top_button, XmNactivateCallback, top_pushed, NULL);

    XtVaGetValues(top_button, XmNbackground, &color, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;

    XpmCreatePixmapFromData(dsp, win, top_arm, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(top_button, XmNlabelType, XmPIXMAP,
		  XmNarmPixmap, image,
		  NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;
    XpmCreatePixmapFromData(dsp, win, top_dis, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(top_button, XmNlabelInsensitivePixmap, image, NULL);

    next_button = XtVaCreateManagedWidget("nextButton",
				     xmPushButtonGadgetClass, button_row,
					  XmNsensitive, False,
					  XmNshowAsDefault, False,
					  NULL);
    XtAddCallback(next_button, XmNactivateCallback, next_pushed, NULL);

    XtVaGetValues(next_button, XmNbackground, &color, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;

    XpmCreatePixmapFromData(dsp, win, next_arm, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(next_button, XmNlabelType, XmPIXMAP,
		  XmNarmPixmap, image,
		  NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;
    XpmCreatePixmapFromData(dsp, win, next_dis, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(next_button, XmNlabelInsensitivePixmap, image, NULL);

    home_button = XtVaCreateManagedWidget("homeButton",
				     xmPushButtonGadgetClass, button_row,
					  XmNsensitive, True,
					  XmNshowAsDefault, False,
					  NULL);
    XtAddCallback(home_button, XmNactivateCallback, home_pushed, NULL);

    XtVaGetValues(home_button, XmNbackground, &color, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;

    XpmCreatePixmapFromData(dsp, win, home, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(home_button, XmNlabelType, XmPIXMAP,
		  XmNarmPixmap, image,
		  NULL);

    XtVaCreateManagedWidget(" ", xmLabelWidgetClass, button_row, NULL);

    search_button = XtVaCreateManagedWidget("searchButton",
				     xmPushButtonGadgetClass, button_row,
					    XmNsensitive, True,
					    XmNshowAsDefault, False,
					    NULL);
    XtAddCallback(search_button, XmNactivateCallback, search_pushed, NULL);

    XtVaGetValues(search_button, XmNbackground, &color, NULL);

    symbols[0].name = NULL;
    symbols[0].value = "none";
    symbols[0].pixel = color;

    attributes.colorsymbols = symbols;
    attributes.numsymbols = 1;
    attributes.valuemask = XpmColorSymbols;

    XpmCreatePixmapFromData(dsp, win, search, &image, NULL, &attributes);
    attributes.colorsymbols[0].pixel = color;

    XtVaSetValues(search_button, XmNlabelType, XmPIXMAP,
		  XmNarmPixmap, image,
		  NULL);

    XtVaCreateManagedWidget(" ", xmLabelWidgetClass, button_row, NULL);

    XtManageChild(button_row);
    location_row = XtVaCreateWidget("locationRow", xmRowColumnWidgetClass, form,
				    XmNorientation, XmHORIZONTAL,
				    XmNtopAttachment, XmATTACH_WIDGET,
				    XmNtopWidget, button_row,
				    XmNleftAttachment, XmATTACH_FORM,
				    XmNrightAttachment, XmATTACH_FORM,
				    XmNadjustLast, False,
				    XmNresizeHeight, True,
				    XmNresizeWidth, True,
				    NULL);

    XtVaCreateManagedWidget(" ", xmLabelWidgetClass, location_row, NULL);

    XtVaCreateManagedWidget("Location:", xmLabelWidgetClass, location_row,
			    NULL);

    location_text = XtVaCreateManagedWidget("locationText",
				    xmTextFieldWidgetClass, location_row,
					    XmNeditable, False,
				     XmNrightAttachment, XmATTACH_WIDGET,
					    XmNresizeWidth, True,
					 XmNcursorPositionVisible, False,
					    XmNcolumns, 64,
					    NULL);
    XmTextFieldSetString(location_text, "No Info File Loaded");

    XtVaCreateManagedWidget(" ", xmLabelWidgetClass, location_row, NULL);

    XtManageChild(location_row);

    separator = XtVaCreateManagedWidget("separator",
					xmSeparatorGadgetClass, form,
					XmNwidth, 8,
				   XmNseparatorType, XmSHADOW_ETCHED_OUT,
					XmNleftAttachment, XmATTACH_FORM,
					XmNrightAttachment, XmATTACH_FORM,
					XmNtopAttachment, XmATTACH_WIDGET,
					XmNtopWidget, location_row,
					NULL);
    /*
     * Create the text area for displaying node contents.
     */

    i = 0;
    XtSetArg(args[i], XmNrows, 40);
    i++;
    XtSetArg(args[i], XmNcolumns, 80);
    i++;
    XtSetArg(args[i], XmNborderWidth, 4);
    i++;
    XtSetArg(args[i], XmNeditable, False);
    i++;
    XtSetArg(args[i], XmNeditMode, XmMULTI_LINE_EDIT);
    i++;
    XtSetArg(args[i], XmNscrollHorizontal, False);
    i++;
    XtSetArg(args[i], XmNscrollLeftSide, True);
    i++;
    XtSetArg(args[i], XmNwordWrap, True);
    i++;
    XtSetArg(args[i], XmNcursorPositionVisible, False);
    i++;
    XtSetArg(args[i], XmNautoShowCursorPosition, True);
    i++;
    XtSetArg(args[i], XmNtopAttachment, XmATTACH_WIDGET);
    i++;
    XtSetArg(args[i], XmNtopWidget, separator);
    i++;
    XtSetArg(args[i], XmNleftAttachment, XmATTACH_FORM);
    i++;
    XtSetArg(args[i], XmNrightAttachment, XmATTACH_FORM);
    i++;
    XtSetArg(args[i], XmNbottomAttachment, XmATTACH_FORM);
    i++;
    XtSetArg(args[i], XmNborderWidth, 0);
    i++;

    node_text = XmCreateScrolledText(form, "nodeText", args, i);
    XtAddCallback(node_text, XmNmotionVerifyCallback, text_cb, node_text);

    XtManageChild(node_text);
    XtManageChild(main_window);
    XtManageChild(form);
    XtRealizeWidget(top_level);

    /* Communicate the final file path to whoever needs it in file_ops
     * globally! Make shure it was really allocated.
     */
    file_path = strdup(app_data.path);
    page_stack = NULL;
    fwrd_stack = NULL;

    first_page = lookup_node_in_file("Top", app_data.file);
    if (first_page)
	display_page(first_page);

    /* Spin until we get sick */
    XtAppMainLoop(app);

    return 0;			/* hope not */
}
