/*
 * (C) 1996 by Marcin Dalecki <sub994.sub.uni-goettingen.de>
 *
 * This file is part of the Motif Info browser.
 *
 * The GNU Info widget is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <string.h>

#include "parse.h"

void normalize_white_space(char **text)
{
    char *n;
    int i;

    while (**text && isspace(**text))
	++ * text;

    while ((n = strstr(*text, "  "))) {
	for (i = 0; i < strlen(n); i++)
	    n[i] = n[i + 1];
    }
}

/*
 * This routine parses the information in an indirection file
 */
struct top *parse_top(char *text)
{
    char *buf;
    char *buf_dup;
    char *buf_tail;
    char *token;

    struct top *top;
    struct indirect *indirect;
    struct node *node;

    if (!text)
	return NULL;

    text = strstr(text, "\x1f\n");

    if (!text)
	return NULL;		/* no heading in file */

    text += 2;

    if (text != strstr(text, "Indirect:\n"))
	return NULL;
    else
	text += strlen("Indirect:\n");

    /* We are now at the top of the indirection table.
     */

    top = NULL;
    buf_dup = buf = strdup(text);
    buf_tail = buf + strlen(buf);

    while (buf < buf_tail) {
	struct indirect *curr;
	char *dp;
	int length;

	if (dp = strchr(buf, '\n'))
	    *dp = '\0';

	if (!strcmp(buf, "\x1f"))
	    break;

	length = strlen(buf);

	dp = strchr(buf, ':');

	if (dp)
	    *dp = '\0';

	curr = (struct indirect *) malloc(sizeof(struct indirect));

	curr->next = NULL;
	curr->file = strdup(buf);
	curr->tag = atoi(dp + 1);
	if (!top) {
	    top = (struct top *) malloc(sizeof(struct top));
	    top->node = NULL;
	    top->indirect = curr;
	} else
	    indirect->next = curr;
	indirect = curr;

	buf += length + 1;
    }
    free(buf_dup);

    /*
     * Now parse the tag table (if any);
     */
    if (top && (text = strstr(text, "\x1f\nTag Table:\n(Indirect)\n"))) {
	text += strlen("\x1f\nTag Table:\n(Indirect)\n");
	buf_dup = buf = strdup(text);
	buf_tail = buf + strlen(buf);

	while (buf < buf_tail) {
	    struct node *curr;
	    char *dp;
	    int length;

	    if (dp = strchr(buf, '\n'))
		*dp = '\0';


	    if (!strcmp(buf, "\x1f"))
		break;

	    length = strlen(buf);

	    dp = strchr(buf, '\x7f');

	    if (dp)
		*dp = '\0';

	    if (token = strstr(buf, "Node: "))
		token += strlen("Node: ");

	    curr = (struct node *) malloc(sizeof(struct node));
	    curr->next = NULL;
	    curr->node = strdup(token);
	    curr->tag = atoi(dp + 1);
	    curr->file = NULL;
	    /* lookup filename; 
	     * we are assuming here that the indirection tabe is sorted in
	     * lower to upper order.
	     */
	    if (top->indirect) {
		struct indirect *last;

		indirect = top->indirect;
		last = indirect;
		while (indirect && (curr->tag >= indirect->tag)) {
		    last = indirect;
		    indirect = indirect->next;
		}
		if (last)
		    curr->file = last->file;
	    }
	    if (!top->node)
		top->node = curr;
	    else
		node->next = curr;
	    node = curr;

	    buf += length + 1;
	}
	free(buf_dup);
    }
#if 0
    debug_print_top(top);
#endif
    return top;
}

/*
 * This routine parses the menuentries in a page
 */

static struct menu *parse_menu(char *text)
{
    char *buf;
    char *buf_head;
    char *buf_tail;
    struct menu *first;
    struct menu *last;
    int length;

    if (!text)
	return NULL;

    last = first = NULL;

    /*
     * Split up into lines...
     */
    buf_head = buf = strdup(text);
    buf_tail = buf_head + strlen(buf_head);
    for (; buf < buf_tail; buf += length + 1) {
	struct menu *menu;
	char *name;
	char *node;
	char *end;
	char *dot;
	char *tail;
	char *token;

	tail = strchr(buf, '\n');
	if (tail)
	    *tail = '\0';

	length = strlen(buf);

	if (strlen(buf) <= 3 || strncmp(buf, "* ", 2))
	    continue;
	token = buf + strlen("* ");

	name = node = NULL;
	/*
	 * First try if it is a double node menu entry.
	 * Thereafter try if it is a menu to a node different then
	 * it's name.
	 */
	if ((end = strstr(token, "::"))) {
	    *end = '\0';
	    name = strdup(token);
	    node = strdup(token);
	} else {
	    if (!(end = strstr(token, ": ")))
		continue;
	    *end = '\0';
	    end += strlen(": ");

	    if (!strcmp(token, "menu") || !strcmp(token, "Menu"))
		continue;

	    if (*end == '(') {
		char *tmp;

		if (!(tmp = strchr(end, ')')))
		    continue;
		if (tmp[1] != '.')
		    continue;
		dot = tmp + 1;
	    } else if (!(dot = strchr(end, '.')))
		continue;
	    *dot = '\0';
	    name = strdup(token);
	    node = strdup(end);
	}

	/*
	 * Gotcha!
	 */
	menu = (struct menu *) malloc(sizeof(struct menu));
	menu->next = NULL;
	menu->name = name;
	while (node && isspace(*node))
	    node++;
	menu->node = node;
	if (!first)
	    last = first = menu;
	else if (last)
	    last->next = menu;
	last = menu;
    }

    free(buf_head);
    return first;
}


/*
 * This routine parses the notes in a page
 */

static struct menu *parse_notes(char *text)
{
    char *buf;
    char *buf_head;
    char *buf_tail;
    struct menu *first;
    struct menu *last;
    int simple;

    if (!text)
	return NULL;

    last = first = NULL;

    buf_head = buf = strdup(text);
    buf_tail = buf_head + strlen(buf_head);

    /*
     * First replace all newlines with spaces.
     */

    while (*buf != '\0') {
	if (*buf == '\n')
	    *buf = ' ';
	++buf;
    }
    buf = buf_head;

    /*
     * Search for notes.
     */

    for (;;) {
	struct menu *xref;
	char *n;
	char *N;
	char *next;
	char *name;
	char *node;
	int i;

	n = strstr(buf, "*Note ");
	N = strstr(buf, "*note ");

	if (!n && !N)
	    break;

	if (!n)
	    next = N;
	else if (!N)
	    next = n;
	else if (n < N)
	    next = n;
	else
	    next = N;
	buf = next + strlen("*Note ");

	if (!(next = strchr(buf, ':')))
	    break;

	if ((next[1] != ':') && (next[1] != ' '))
	    continue;

	if (next[1] == ':')
	    simple = 1;
	else
	    simple = 0;
	*next = '\0';

	normalize_white_space(&buf);
	name = strdup(buf);
	if (simple) {
	    node = strdup(name);
	    buf = next + 1;
	} else {
	    buf = next + 2;
	    if (buf >= buf_tail) {
		free(name);
		break;
	    }
	    next = strpbrk(buf, ",.");
	    if (!next) {
		free(name);
		continue;
	    }
	    if (next == strstr(next, ".info")) {
		next += strlen(".info");
		next = strpbrk(next, ",.");
		if (!next) {
		    free(name);
		    continue;
		}
	    } else if (next == strstr(next, "info")) {
		next += strlen("info");
		next = strpbrk(next, ",.");
		if (!next) {
		    free(name);
		    continue;
		}
	    }
	    *next = '\0';

	    normalize_white_space(&buf);
	    node = strdup(buf);
	    buf = next + 1;
	}

	/*
	 * Gotcha! we have found a cross reference.
	 */

	xref = first;
	while (xref) {
	    if (!strcmp(xref->name, name) && !strcmp(xref->node, node)) {
		free(name);
		free(node);
		break;
	    }
	    xref = xref->next;
	}
	if (!xref) {
	    xref = (struct menu *) malloc(sizeof(struct menu));
	    xref->next = NULL;
	    xref->name = name;
	    xref->node = node;
	    if (!first)
		last = first = xref;
	    else if (last)
		last->next = xref;
	    last = xref;
	}
    };

    free(buf_head);
    return first;
}

/*
 * This routine returns a list of nodes contained in an actual data file.
 */
struct page *parse_file(char *file, char *text)
{
    struct page *first;
    struct page *last;
    char *buf;
    char *buf_head;
    char *buf_tail;

    if (!text)
	return NULL;

    if (!text)
	return NULL;		/* no heading in file */

    if (!(text = strchr(text, '\x1f')))
	return NULL;
    text += 1;

    last = first = NULL;

    buf_head = buf = strdup(text);
    buf_tail = buf + strlen(buf);
    while (buf < buf_tail) {
	struct page *page;
	char *head;
	char *tail;
	int length;

	tail = strchr(buf, '\x1f');
	if (tail)
	    *tail = '\0';

	length = strlen(buf);

	page = (struct page *) malloc(sizeof(struct page));
	page->next = NULL;
	page->file = strdup(file);
	page->File = NULL;
	page->Node = NULL;
	page->Next = NULL;
	page->Prev = NULL;
	page->Up = NULL;
	page->text = NULL;
	page->menu = NULL;
	page->note = NULL;
	if (!first)
	    last = first = page;
	else if (last)
	    last->next = page;
	last = page;

	/* 
	 * Parse the head of the page
	 */

	head = strchr(buf + 1, '\n');
	if (head) {
	    char *end;
	    char *begin;
	    char *cont;

	    *head = '\0';
	    cont = head + 1;
	    if ((begin = strstr(buf, "\nFile: "))
		&& ((end = strchr(begin + strlen("\nFile: "), ','))
		|| (end = strpbrk(begin + strlen("\nFile: "), " \t")))) {
		*end = '\0';
		page->File = strdup(begin + strlen("\nFile: "));
		head = end + 1;
	    }
	    if ((begin = strstr(head, "Node: "))
		&& ((end = strchr(begin + strlen("Node: "), ','))
		  || (end = strpbrk(begin + strlen("Node: "), " \t")))) {
		*end = '\0';
		page->Node = strdup(begin + strlen("Node: "));
		head = end + 1;
	    }
	    if ((begin = strstr(head, "Next: "))
		&& ((end = strchr(begin + strlen("Next: "), ','))
		  || (end = strpbrk(begin + strlen("Next: "), " \t")))) {
		*end = '\0';
		page->Next = strdup(begin + strlen("Next: "));
		head = end + 1;
	    }
	    if ((begin = strstr(head, "Prev: "))
		&& ((end = strchr(begin + strlen("Prev: "), ','))
		  || (end = strpbrk(begin + strlen("Prev: "), " \t")))) {
		*end = '\0';
		page->Prev = strdup(begin + strlen("Prev: "));
		head = end + 1;
	    }
	    if ((begin = strstr(head, "Up: "))) {
		page->Up = strdup(begin + strlen("Up: "));
	    }
	    page->text = strdup(cont);
	} else
	    page->text = NULL;
#if 0
	printf("File: %s\nNode: %s\nPrev %s\nNext %s\nUp: %s\n\n",
	       page->File, page->Node, page->Prev, page->Next, page->Up);
#endif
	/*
	 * Parse the menu entries of the node.
	 */
	page->menu = parse_menu(page->text);
	page->note = parse_notes(page->text);
	buf += length + 1;
    }
    free(buf_head);

    return first;
}

/*
 * Routines usefull for debugging output.
 */

void debug_print_top(struct top *top)
{
    struct indirect *indirect;
    struct node *node;

    if (!top)
	return;

    puts("\nIndirections:");
    for (indirect = top->indirect; indirect; indirect = indirect->next)
	printf("\tfile: %s, tag: %d\n", indirect->file, indirect->tag);
    puts("\nNode indirections:");
    for (node = top->node; node; node = node->next)
	printf("\tnode: %s, file: %s tag: %d\n",
	       node->node, node->file, node->tag);

}
void debug_print_page(struct page *page)
{
    if (!page)
	return;
    puts("BEGIN PAGE");
    printf("\tFile: [%s]\n", page->File);
    printf("\tNode: [%s]\n", page->Node);
    printf("\tNext: [%s]\n", page->Next);
    printf("\tPrev: [%s]\n", page->Prev);
    printf("\tUp: [%s]\n", page->Up);
    if (page->menu) {
	struct menu *menu;
	puts("\tBEGIN MENU");
	for (menu = page->menu; menu; menu = menu->next)
	    printf("\t\tName: [%s], Node: [%s]\n", menu->name, menu->node);
	puts("\tEND MENU");
    }
    puts(page->text);
    puts("END PAGE");
}
