/*
 * (C) 1996 by Marcin Dalecki <sub994.sub.uni-goettingen.de>
 *
 * This file is part of the Motif Info browser.
 *
 * The GNU Info widget is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <Xm/MessageB.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/TextF.h>

#include "file_ops.h"
#include "display.h"

void error_dialog(char *message)
{
    Widget dialog;
    Arg args[5];
    int i;

    XmString msg;
    msg = XmStringCreateSimple(message);
    i = 0;
    XtSetArg(args[i], XmNmessageString, msg);
    i++;

    dialog = (Widget) XmCreateErrorDialog(top_level,
					  "dirFailedDialog", args, i);
    XtUnmanageChild((void *) XmMessageBoxGetChild(dialog,
						  XmDIALOG_HELP_BUTTON));
    XtUnmanageChild((void *) XmMessageBoxGetChild(dialog,
						XmDIALOG_CANCEL_BUTTON));
    XmStringFree(msg);
    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);
}


static Widget search_field;
static Widget search_dialog;


static struct {
    int forward;
    int ignorecase;
} search_data;

static void case_toggled(widget, client_data, call_data)
Widget widget;
XtPointer client_data;
XtPointer call_data;
{
    XmToggleButtonCallbackStruct *state =
    (XmToggleButtonCallbackStruct *) call_data;

    printf("%s: %s\n", XtName(widget), state->set ? "on" : "off");
    if (state->set)
	search_data.ignorecase = False;
    else
	search_data.ignorecase = True;
}

static void direction_toggled(widget, client_data, call_data)
Widget widget;
XtPointer client_data;
XtPointer call_data;
{
    XmToggleButtonCallbackStruct *state =
    (XmToggleButtonCallbackStruct *) call_data;

    if (state->set)
	search_data.forward = False;
    else
	search_data.forward = True;
}

static void find_callback(Widget w,
			  XtPointer client_data,
			  XtPointer call_data)
{
    String needle;
    XmTextPosition found;
    XmTextPosition current;
    XmTextDirection direction;

    needle = XmTextFieldGetString(search_field);
    if (!needle || !strcmp(needle, "")) {
	error_dialog("Nothing to search for.");
	XtFree(needle);
	return;
    }
    if (!node_text) {
	error_dialog("Nothing to search in.");
	XtFree(needle);
	return;
    }
    if (search_data.forward)
	direction = XmTEXT_FORWARD;
    else
	direction = XmTEXT_BACKWARD;

    current = XmTextGetInsertionPosition(node_text);

    if (!XmTextFindString(node_text, current, needle, direction, &found)) {
	Widget dialog;
	Arg args[5];
	int i;

	XmString msg;
	msg = XmStringCreateSimple("Search string not found!");
	i = 0;
	XtSetArg(args[i], XmNmessageString, msg);
	i++;

	dialog = (Widget) XmCreateInformationDialog(top_level,
					  "searchFailedDialog", args, i);
	XtUnmanageChild((void *) XmMessageBoxGetChild(dialog,
						  XmDIALOG_HELP_BUTTON));
	XtUnmanageChild((void *) XmMessageBoxGetChild(dialog,
						XmDIALOG_CANCEL_BUTTON));
	XmStringFree(msg);
	XtManageChild(dialog);
	XtPopup(XtParent(dialog), XtGrabNone);
	XtFree(needle);
	return;
    } else {
	XmTextSetInsertionPosition(node_text, found);
	/* And now highlight the found string.
	 * thereafter try to set the insertion position beyoind it!
	 */
	XmTextSetHighlight(node_text, found, found + strlen(needle),
			   XmHIGHLIGHT_SELECTED);
	XmTextSetInsertionPosition(node_text, found + strlen(needle));

    }
    XtFree(needle);
}

static void clear_callback(Widget w,
			   XtPointer client_data,
			   XtPointer call_data)
{
    XmTextFieldSetString(search_field, "");
    XmTextSetInsertionPosition(node_text, 0);
}

static void close_callback(Widget w,
			   XtPointer client_data,
			   XtPointer call_data)
{
    /*
     * *FIXME*
     * How to reset the hightlight's????
     */
    XtUnmanageChild((Widget) client_data);
    XmTextSetInsertionPosition(node_text, 0);
}


void search_pushed(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    Widget form, label, find, clear, close, case_toggle, direction_toggle;
    XmString str;

    if (!search_dialog) {
	search_dialog = XmCreateTemplateDialog(top_level, "NONAME", NULL, 0);

	search_data.forward = True;
	search_data.ignorecase = True;

	XmTextSetInsertionPosition(node_text, 0);
	XtVaSetValues(search_dialog,
		      XmNdefaultButton, find,
		      XmNautoUnmanage, False,
		      NULL);
	find = XtCreateManagedWidget("Find", xmPushButtonWidgetClass,
				     search_dialog, NULL, 0);


	clear = XtCreateManagedWidget("Clear", xmPushButtonWidgetClass,
				      search_dialog, NULL, 0);
	close = XtCreateManagedWidget("Close", xmPushButtonWidgetClass,
				      search_dialog, NULL, 0);


	XtAddCallback(find, XmNactivateCallback, find_callback, NULL);
	XtAddCallback(clear, XmNactivateCallback, clear_callback, NULL);
	XtAddCallback(close, XmNactivateCallback, close_callback,
		      search_dialog);

	/* Create a manager widget to handle the dialog layout.
	 */

	form = XtVaCreateManagedWidget("form",
				       xmFormWidgetClass, search_dialog,
				       XmNorientation, XmHORIZONTAL,
				       XmNnumColumns, 2,
				       XmNtopAttachment, XmATTACH_FORM,
				       XmNleftAttachment, XmATTACH_FORM,
				       XmNrightAttachment, XmATTACH_FORM,
				       NULL);

	search_field = XtVaCreateManagedWidget("find",
					    xmTextFieldWidgetClass, form,
					       XmNcolumns, 44,
					       XmNeditable, True,
				       XmNrightAttachment, XmATTACH_FORM,
					 XmNtopAttachment, XmATTACH_FORM,
					       NULL);

	XtVaSetValues(search_dialog, XmNinitialFocus, search_field, NULL);

	label = XtVaCreateManagedWidget("Find:", xmLabelWidgetClass, form,
					XmNleftAttachment, XmATTACH_FORM,
			      XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
					XmNtopWidget, search_field,
			   XmNbottomAttachment, XmATTACH_OPPOSITE_WIDGET,
					XmNbottomWidget, search_field,
				     XmNrightAttachment, XmATTACH_WIDGET,
					XmNrightWidget, search_field,
					NULL);


	str = XmStringCreateLocalized("Case Sensitive");

	case_toggle = XtVaCreateManagedWidget("caseToggle",
					 xmToggleButtonWidgetClass, form,
					      XmNlabelString, str,
					      XmNshowAsDefault, True,
			     XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET,
					      XmNleftWidget, search_field,
				       XmNtopAttachment, XmATTACH_WIDGET,
					      XmNtopWidget, search_field,
					      NULL);

	XmStringFree(str);
	XtAddCallback(case_toggle, XmNvalueChangedCallback,
		      case_toggled, NULL);

	str = XmStringCreateLocalized("Find Backwards");
	direction_toggle = XtVaCreateManagedWidget("findToggle",
					 xmToggleButtonWidgetClass, form,
						   XmNlabelString, str,
						   XmNshowAsDefault, True,
				      XmNleftAttachment, XmATTACH_WIDGET,
					      XmNleftWidget, case_toggle,
						   XmNleftOffset, 32,
				       XmNtopAttachment, XmATTACH_WIDGET,
					      XmNtopWidget, search_field,
						   NULL);
	XmStringFree(str);
	XtAddCallback(direction_toggle, XmNvalueChangedCallback,
		      direction_toggled, NULL);


    }
    XtManageChild(search_dialog);

    XtPopup(XtParent(search_dialog), XtGrabNone);
}
